# ----------------------------------------------------------
# NDN Hydra Client
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import asyncio
from argparse import ArgumentParser, Namespace
import logging
from ndn.app import NDNApp
from ndn.encoding import Name, FormalName
import sys
import os
import pkg_resources
from ndn_hydra.client.functions import *

def parse_hydra_cmd_opts() -> Namespace:
    def interpret_version() -> None:
        set = True if "-v" in sys.argv else False
        if set and (len(sys.argv)-1 < 2):
            try: print("ndn-hydra " + pkg_resources.require("ndn-hydra")[0].version)
            except pkg_resources.DistributionNotFound: print("ndn-hydra source,undetermined")
            sys.exit(0)
    def interpret_help() -> None:
        set = True if "-h" in sys.argv else False
        if set:
            if (len(sys.argv)-1 < 2):
                print("usage: ndn-hydra-client [-h] [-v] {insert,delete,fetch,query} ...")
                print("    ndn-hydra-client: a client made specifically for hydra, the NDN distributed repo.")
                print("    ('python3 ./examples/client.py' instead of 'ndn-hydra-client' if from source.)")
                print("")
                print("* informational args:")
                print("  -h, --help                      |   shows this help message and exits.")
                print("  -v, --version                   |   shows the current version and exits.")
                print("")
                print("* function 'insert':")
                print("     usage: ndn-hydra-client insert -r REPO -f FILENAME -p PATH [-c COPIES]")
                print("     required args:")
                print("        -r, --repoprefix REPO     |   a proper name of the repo prefix.")
                print("        -f, --filename FILENAME   |   a proper name for the input file.")
                print("        -p, --path PATH           |   path of the file desired to be the input i.e. input path.")
                print("     optional args:")
                print("        -c, --copies COPIES       |   number of copies for files, default 2.")
                print("")
                print("* function 'delete':")
                print("     usage: ndn-hydra-client delete -r REPO -f FILENAME")
                print("     required args:")
                print("        -r, --repoprefix REPO     |   a proper name of the repo prefix.")
                print("        -f, --filename FILENAME   |   a proper name for selected file.")
                print("")
                print("* function 'fetch':")
                print("     usage: ndn-hydra-client fetch -r REPO -f FILENAME [-p PATH]")
                print("     required args:")
                print("        -r, --repoprefix REPO     |   a proper name of the repo prefix.")
                print("        -f, --filename FILENAME   |   a proper name for desired file.")
                print("     optional args:")
                print("        -p, --path PATH           |   path for the file to be placed i.e. output path.")
                print("")
                print("* function 'query':")
                print("     usage: ndn-hydra-client query -r REPO -q QUERY [-s SESSIONID]")
                print("     required args:")
                print("        -r, --repoprefix REPO     |   a proper name of the repo prefix.")
                print("        -q, --query QUERY         |   the type of query desired.")
                print("     optional args:")
                print("        -s, --sessionid SESSIONID |   certain sessionid-node targeted for query, default closest node.")
                print("")
                print("Thank you for using hydra.")
            sys.exit(0)
    # Command Line Parser
    parser = ArgumentParser(prog="ndn-hydra-client",add_help=False,allow_abbrev=False)
    parser.add_argument("-h","--help",action="store_true",dest="help",default=False,required=False)
    parser.add_argument("-v","--version",action="store_true",dest="version",default=False,required=False)
    subparsers = parser.add_subparsers(dest="function",required=True)

    # Define All Subparsers
    insertsp = subparsers.add_parser('insert',add_help=False)
    insertsp.add_argument("-r","--repoprefix",action="store",dest="repo",required=True)
    insertsp.add_argument("-f","--filename",action="store",dest="filename",required=True)
    insertsp.add_argument("-p","--path",action="store",dest="path",required=True)
    insertsp.add_argument("-c","--copies",action="store",dest="copies",required=False,default=2,type=int,nargs=None)

    deletesp = subparsers.add_parser('delete',add_help=False)
    deletesp.add_argument("-r","--repoprefix",action="store",dest="repo",required=True)
    deletesp.add_argument("-f","--filename",action="store",dest="filename",required=True)

    fetchsp = subparsers.add_parser('fetch',add_help=False)
    fetchsp.add_argument("-r","--repoprefix",action="store",dest="repo",required=True)
    fetchsp.add_argument("-f","--filename",action="store",dest="filename",required=True)
    fetchsp.add_argument("-p","--path",action="store",dest="path",default="./fetchedHydraFile", required=False)

    querysp = subparsers.add_parser('query',add_help=False)
    querysp.add_argument("-r","--repoprefix",action="store",dest="repo",required=True)
    querysp.add_argument("-q","--query",action="store",dest="query",required=True)
    querysp.add_argument("-s","--sessionid",action="store",dest="sessionid",default=None, required=False)

    # Interpret Informational Arguments
    interpret_version()
    interpret_help()

    # Getting all Arguments
    vars = parser.parse_args()

    # Configure Arguments
    if vars.function == "insert":
        if not os.path.isfile(vars.path):
            print('Error: path specified is not an actual file. Unable to insert.')
            sys.exit()
        if vars.copies < 2:
            print('Error: insufficient number of copies, must be 2 or above.')
            sys.exit()
    return vars

class HydraClient():
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
        self.cinsert = HydraInsertClient(app, client_prefix, repo_prefix)
        self.cdelete = HydraDeleteClient(app, client_prefix, repo_prefix)
        self.cfetch = HydraFetchClient(app, client_prefix, repo_prefix)
        self.cquery = HydraQueryClient(app, client_prefix, repo_prefix)
    async def insert(self, file_name: FormalName, desired_copies: int, path: str) -> bool:
        return await self.cinsert.insert_file(file_name, desired_copies, path);
    async def delete(self, file_name: FormalName) -> bool:
        return await self.cdelete.delete_file(file_name);
    async def fetch(self, file_name: FormalName, local_filename: str = None, overwrite: bool = False) -> None:
        return await self.cfetch.fetch_file(file_name, local_filename, overwrite)
    async def query(self, query: Name, sid: str=None) -> None:
        return await self.cquery.send_query(query, sid)

async def run_hydra_client(app: NDNApp, args: Namespace) -> None:
  repo_prefix = Name.from_str(args.repo)
  client_prefix = Name.from_str("/client")
  filename = None
  desired_copies = 2
  if hasattr(args, 'copies'):
    desired_copies = args.copies
  client = HydraClient(app, client_prefix, repo_prefix)

  if args.function != "query":
      filename = Name.from_str(args.filename)

  if args.function == "insert":
    await client.insert(filename, desired_copies, args.path)
    print("Client finished Insert Command!")
    await asyncio.sleep(60)

  elif args.function == "delete":
    await client.delete(filename)
    print("Client finished Delete Command!")

  elif args.function == "fetch":
    await client.fetch(filename, args.path, True)
    print("Client finished Fetch Command!")

  elif args.function == "query":
    await client.query(Name.from_str(str(args.query)), args.sessionid)
    print("Client finished Query Command!")

  else:
    print("Not Implemented Yet / Unknown Command.")

  app.shutdown()

def main() -> None:
    args = parse_hydra_cmd_opts()
    app = NDNApp()
    try:
        app.run_forever(after_start=run_hydra_client(app, args))
    except FileNotFoundError:
        print('Error: could not connect to NFD.')
        sys.exit()

if __name__ == "__main__":
    sys.exit(main())



# ----------------------------------------------------------
# NDN Hydra Remove Group Message
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

from typing import Callable
from ndn.encoding import *
from ndn_hydra.repo.global_view.global_view import GlobalView
from ndn_hydra.repo.repo_messages.message_base import MessageBodyBase

class RemoveMessageBodyTypes:
    SESSION_ID = 83
    NODE_NAME = 84
    EXPIRE_AT = 85
    FAVOR = 86

    INSERTION_ID = 90

class RemoveMessageBodyTlv(TlvModel):
    session_id = BytesField(RemoveMessageBodyTypes.SESSION_ID)
    node_name = BytesField(RemoveMessageBodyTypes.NODE_NAME)
    expire_at = UintField(RemoveMessageBodyTypes.EXPIRE_AT)
    favor = BytesField(RemoveMessageBodyTypes.FAVOR)
    insertion_id = BytesField(RemoveMessageBodyTypes.INSERTION_ID)

class RemoveMessageBody(MessageBodyBase):
    def __init__(self, nid:str, seq:int, raw_bytes:bytes):
        super(RemoveMessageBody, self).__init__(nid, seq)
        self.message_body = RemoveMessageBodyTlv.parse(raw_bytes)

    async def apply(self, global_view: GlobalView, fetch_file: Callable, svs, config):
        session_id = self.message_body.session_id.tobytes().decode()
        node_name = self.message_body.node_name.tobytes().decode()
        expire_at = self.message_body.expire_at
        favor = float(self.message_body.favor.tobytes().decode())
        insertion_id = self.message_body.insertion_id.tobytes().decode()
        val = "[MSG][REMOVE]  iid={iid}".format(
            sid=session_id,
            iid=insertion_id
        )
        self.logger.info(val)
        # if insertion
        insertion = global_view.get_insertion(insertion_id)
        if (insertion == None) or (insertion['is_deleted'] == True):
            # add store to pending_stores
            self.logger.warning('nothing to remove')
        else:
            global_view.delete_insertion(insertion_id)
            # TODO: remove from data_storage
        # update session
        global_view.update_session(session_id, node_name, expire_at, favor, self.seq)
        return


# ----------------------------------------------------------
# NDN Hydra Add Group Message
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

from typing import Callable
from ndn.encoding import *
import time
from ndn_hydra.repo.global_view.global_view import GlobalView
from ndn_hydra.repo.repo_messages.message_base import MessageBodyBase
from ndn_hydra.repo.repo_messages.store import StoreMessageBodyTlv

class AddMessageBodyTypes:
    SESSION_ID = 83
    NODE_NAME = 84
    EXPIRE_AT = 85
    FAVOR = 86

    INSERTION_ID = 90
    FILE = 91
    DESIRED_COPIES = 92
    PACKETS = 93
    DIGEST = 94
    SIZE = 95
    SEQUENCE_NUMBER = 96
    FETCH_PATH = 97
    IS_STORED_BY_ORIGIN = 98

    BACKUP = 100
    BACKUP_SESSION_ID = 101
    BACKUP_NONCE = 102

class FileTlv(TlvModel):
    file_name = NameField()
    desired_copies = UintField(AddMessageBodyTypes.DESIRED_COPIES, default=3)
    packets = UintField(AddMessageBodyTypes.PACKETS)
    digests = RepeatedField(BytesField(AddMessageBodyTypes.DIGEST))
    size = UintField(AddMessageBodyTypes.SIZE)

class FetchPathTlv(TlvModel):
    prefix = NameField()

class BackupTlv(TlvModel):
    session_id = BytesField(AddMessageBodyTypes.BACKUP_SESSION_ID)
    nonce = BytesField(AddMessageBodyTypes.BACKUP_NONCE)

class AddMessageBodyTlv(TlvModel):
    session_id = BytesField(AddMessageBodyTypes.SESSION_ID)
    node_name = BytesField(AddMessageBodyTypes.NODE_NAME)
    expire_at = UintField(AddMessageBodyTypes.EXPIRE_AT)
    favor = BytesField(AddMessageBodyTypes.FAVOR)
    insertion_id = BytesField(AddMessageBodyTypes.INSERTION_ID)
    file = ModelField(AddMessageBodyTypes.FILE, FileTlv)

    sequence_number = UintField(AddMessageBodyTypes.SEQUENCE_NUMBER)
    fetch_path = ModelField(AddMessageBodyTypes.FETCH_PATH, FetchPathTlv)
    is_stored_by_origin = UintField(AddMessageBodyTypes.IS_STORED_BY_ORIGIN)
    backup_list = RepeatedField(ModelField(AddMessageBodyTypes.BACKUP, BackupTlv))

class AddMessageBody(MessageBodyBase):
    def __init__(self, nid:str, seq:int, raw_bytes:bytes):
        super(AddMessageBody, self).__init__(nid, seq)
        self.message_body = AddMessageBodyTlv.parse(raw_bytes)

    async def apply(self, global_view: GlobalView, fetch_file: Callable, svs, config):
        session_id = self.message_body.session_id.tobytes().decode()
        node_name = self.message_body.node_name.tobytes().decode()
        expire_at = self.message_body.expire_at
        favor = float(self.message_body.favor.tobytes().decode())
        insertion_id = self.message_body.insertion_id.tobytes().decode()
        file = self.message_body.file
        file_name = file.file_name
        desired_copies = file.desired_copies
        packets = file.packets
        digests = file.digests
        size = file.size
        sequence_number = self.message_body.sequence_number
        fetch_path = self.message_body.fetch_path.prefix
        is_stored_by_origin = False if (self.message_body.is_stored_by_origin == 0) else True
        backups = self.message_body.backup_list
        backup_list = []
        bak = ""
        for backup in backups:
            backup_list.append((backup.session_id.tobytes().decode(), backup.nonce.tobytes().decode()))
            bak = bak + backup.session_id.tobytes().decode() + ","
        val = "[MSG][ADD]     sid={sid};iid={iid};file={fil};cop={cop};pck={pck};siz={siz};seq={seq};slf={slf};bak={bak}".format(
            sid=session_id,
            iid=insertion_id,
            fil=Name.to_str(file_name),
            cop=desired_copies,
            pck=packets,
            siz=size,
            seq=sequence_number,
            slf=1 if is_stored_by_origin else 0,
            bak=bak
        )
        self.logger.info(val)
        global_view.add_insertion(
            insertion_id,
            Name.to_str(file_name),
            sequence_number,
            size,
            session_id,
            Name.to_str(fetch_path),
            self.seq,
            b''.join(digests),
            packets=packets,
            desired_copies=desired_copies
        )

        global_view.set_backups(insertion_id, backup_list)

        # get pending stores
        copies_needed = desired_copies
        pending_stores = global_view.get_pending_stores(insertion_id)
        for pending_store in pending_stores:
            # data_storage.add_metainfos(insertion_id, Name.to_str(file_name), packets, digests, Name.to_str(fetch_path))
            global_view.store_file(insertion_id, pending_store)
            copies_needed -= 1

        # if I need to store this file
        # if is_stored_by_origin:
        #     copies_needed -= 1
        need_to_store = False
        for i in range(copies_needed):
            backup = backup_list[i]
            if backup[0] == config['session_id']:
                need_to_store = True
                break
        if need_to_store == True:
            fetch_file(insertion_id, Name.to_str(file_name), packets, digests, Name.to_str(fetch_path))

            # from .message import MessageTlv, MessageTypes
            # # generate store msg and send
            # # store tlv
            # expire_at = int(time.time()+(config['period']*2))
            # favor = 1.85
            # store_message_body = StoreMessageBodyTlv()
            # store_message_body.session_id = config['session_id'].encode()
            # store_message_body.node_name = config['node_name'].encode()
            # store_message_body.expire_at = expire_at
            # store_message_body.favor = str(favor).encode()
            # store_message_body.insertion_id = insertion_id.encode()
            # # store msg
            # store_message = MessageTlv()
            # store_message.header = MessageTypes.STORE
            # store_message.body = store_message_body.encode()
            # # apply globalview and send msg thru SVS
            # # next_state_vector = svs.getCore().getStateVector().get(config['session_id']) + 1

            # # global_view.store_file(insertion_id, config['session_id'])
            # svs.publishData(store_message.encode())
            # val = "[MSG][STORE]*  sid={sid};iid={iid}".format(
            #     sid=config['session_id'],
            #     iid=insertion_id
            # )
            # self.logger.info(val)
        # update session
        global_view.update_session(session_id, node_name, expire_at, favor, self.seq)
        return



# ----------------------------------------------------------
# NDN Hydra General Group Message
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

from ndn.encoding import *
from ndn_hydra.repo.repo_messages.add import AddMessageBody
from ndn_hydra.repo.repo_messages.remove import RemoveMessageBody
from ndn_hydra.repo.repo_messages.store import StoreMessageBody
from ndn_hydra.repo.repo_messages.claim import ClaimMessageBody
from ndn_hydra.repo.repo_messages.heartbeat import HeartbeatMessageBody
from ndn_hydra.repo.repo_messages.expire import ExpireMessageBody

class MessageTypes:
    ADD = 1
    REMOVE = 2
    STORE = 3
    CLAIM = 4
    HEARTBEAT = 5
    EXPIRE = 6

class MessageTlv(TlvModel):
    header = UintField(0x80)
    body = BytesField(0x81)

class Message:
    def __init__(self, nid:str, seq:int, raw_bytes:bytes):
        self.nid = nid
        self.seq = seq
        self.message = MessageTlv.parse(raw_bytes)

    def get_message_header(self):
        return self.message.header

    def get_message_body(self):
        message_type = self.message.header
        raw_bytes = self.message.body.tobytes()
        if message_type == MessageTypes.ADD:
            return AddMessageBody(self.nid, self.seq, raw_bytes)
        elif message_type == MessageTypes.REMOVE:
            return RemoveMessageBody(self.nid, self.seq, raw_bytes)
        elif message_type == MessageTypes.STORE:
            return StoreMessageBody(self.nid, self.seq, raw_bytes)
        elif message_type == MessageTypes.CLAIM:
            return ClaimMessageBody(self.nid, self.seq, raw_bytes)
        elif message_type == MessageTypes.HEARTBEAT:
            return HeartbeatMessageBody(self.nid, self.seq, raw_bytes)
        elif message_type == MessageTypes.EXPIRE:
            return ExpireMessageBody(self.nid, self.seq, raw_bytes)
        else:
            return None





# ----------------------------------------------------------
# NDN Hydra Claim Group Message
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import copy
from typing import Callable
import time
from ndn.encoding import *
from ndn_hydra.repo.global_view.global_view import GlobalView
from ndn_hydra.repo.repo_messages.message_base import MessageBodyBase

class ClaimMessageTypes:
    REQUEST = 1
    COMMITMENT = 2

class ClaimMessageBodyTypes:
    SESSION_ID = 83
    NODE_NAME = 84
    EXPIRE_AT = 85
    FAVOR = 86

    INSERTION_ID = 90
    TYPE = 91 # 1=request; 2=commitment
    CLAIMER_SESSION_ID = 92
    CLAIMER_NONCE = 93
    AUTHORIZER_SESSION_ID = 94
    AUTHORIZER_NONCE = 95


class ClaimMessageBodyTlv(TlvModel):
    session_id = BytesField(ClaimMessageBodyTypes.SESSION_ID)
    node_name = BytesField(ClaimMessageBodyTypes.NODE_NAME)
    expire_at = UintField(ClaimMessageBodyTypes.EXPIRE_AT)
    favor = BytesField(ClaimMessageBodyTypes.FAVOR)
    insertion_id = BytesField(ClaimMessageBodyTypes.INSERTION_ID)
    type = UintField(ClaimMessageBodyTypes.TYPE)
    claimer_session_id = BytesField(ClaimMessageBodyTypes.CLAIMER_SESSION_ID)
    claimer_nonce = BytesField(ClaimMessageBodyTypes.CLAIMER_NONCE)
    authorizer_session_id = BytesField(ClaimMessageBodyTypes.AUTHORIZER_SESSION_ID)
    authorizer_nonce = BytesField(ClaimMessageBodyTypes.AUTHORIZER_NONCE)

class ClaimMessageBody(MessageBodyBase):
    def __init__(self, nid:str, seq:int, raw_bytes:bytes):
        super(ClaimMessageBody, self).__init__(nid, seq)
        self.message_body = ClaimMessageBodyTlv.parse(raw_bytes)

    async def apply(self, global_view: GlobalView, fetch_file: Callable, svs, config):
        session_id = self.message_body.session_id.tobytes().decode()
        node_name = self.message_body.node_name.tobytes().decode()
        expire_at = self.message_body.expire_at
        favor = float(self.message_body.favor.tobytes().decode())
        insertion_id = self.message_body.insertion_id.tobytes().decode()
        type = self.message_body.type
        claimer_session_id = self.message_body.claimer_session_id.tobytes().decode()
        claimer_nonce = self.message_body.claimer_nonce.tobytes().decode()
        authorizer_session_id = self.message_body.authorizer_session_id.tobytes().decode()
        authorizer_nonce = self.message_body.authorizer_nonce.tobytes().decode()
        insertion = global_view.get_insertion(insertion_id)
        if type == ClaimMessageTypes.COMMITMENT:
            rank = len(insertion['backuped_bys'])
            val = "[MSG][CLAIM.C] sid={sid};iid={iid}".format(
                sid=claimer_session_id,
                iid=insertion_id
            )
            self.logger.info(val)
            global_view.add_backup(insertion_id, claimer_session_id, rank, claimer_nonce)
        else:
            val = "[MSG][CLAIM.R] sid={sid};iid={iid}".format(
                sid=claimer_session_id,
                iid=insertion_id
            )
            self.logger.info(val)
            if authorizer_session_id == config['session_id']:
                from .message import MessageTlv, MessageTypes
                commit = False
                if (len(insertion['backuped_bys']) == 0) and (insertion['stored_bys'][-1] == config['session_id']) and (authorizer_nonce == insertion['id']):
                    global_view.add_backup(insertion_id, claimer_session_id, 0, claimer_nonce)
                    commit = True
                if (len(insertion['backuped_bys']) > 0) and (insertion['backuped_bys'][-1]['session_id'] == config['session_id']) and (authorizer_nonce == insertion['backuped_bys'][-1]['nonce']):
                    global_view.add_backup(insertion_id, claimer_session_id, len(insertion['backuped_bys']), claimer_nonce)
                    commit = True
                if commit == True:
                    # claim tlv
                    expire_at = int(time.time()+(config['period']*2))
                    favor = 1.85
                    claim_message_body = copy.copy(self.message_body)
                    claim_message_body.session_id = config['session_id'].encode()
                    claim_message_body.node_name = config['node_name'].encode()
                    claim_message_body.expire_at = expire_at
                    claim_message_body.favor = str(favor).encode()
                    claim_message_body.type = ClaimMessageTypes.COMMITMENT
                    # claim msg
                    claim_message = MessageTlv()
                    claim_message.header = MessageTypes.CLAIM
                    claim_message.body = claim_message_body.encode()
                    svs.publishData(claim_message.encode())
                    val = "[MSG][CLAIM.C]*sid={sid};iid={iid}".format(
                        sid=claimer_session_id,
                        iid=insertion_id
                    )
                    self.logger.info(val)
        # update session
        global_view.update_session(session_id, node_name, expire_at, favor, self.seq)
        return


# ----------------------------------------------------------
# NDN Hydra MainLoop
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import asyncio as aio
import logging
import secrets
import time
import random
from typing import Dict, List
from ndn.app import NDNApp
from ndn.encoding import Name, Component
from ndn.types import InterestNack, InterestTimeout
from ndn.svs import SVSync
from ndn_python_repo import Storage, SqliteStorage
from ndn_hydra.repo.global_view.global_view import GlobalView
from ndn_hydra.repo.repo_messages import *
from ndn_hydra.repo.utils.concurrent_fetcher import concurrent_fetcher

class MainLoop:
    def __init__(self, app: NDNApp, config: Dict, global_view: GlobalView, data_storage: Storage):
        self.app = app
        self.config = config
        self.global_view = global_view
        self.data_storage = data_storage
        self.svs_storage = SqliteStorage(self.config['svs_storage_path'])
        self.svs = None
        self.expire_at = 0
        self.logger = logging.getLogger()

    # the main coroutine
    async def start(self):
        self.svs = SVSync(self.app, Name.normalize(self.config['repo_prefix'] + "/group"), Name.normalize(self.config['session_id']), self.svs_missing_callback, storage=self.svs_storage)
        while True:
            await aio.sleep(self.config['period'])
            self.periodic()

    # def __init__(self, app:NDNApp, svs_storage:Storage, session_id:str, node_name:str, svs_cache_others:bool, global_view: GlobalView, config):
    #     self.app = app
    #     self.svs_storage = svs_storage
    #     self.svs_group_prefix = repo_prefix + "/group"
    #     self.svs_node_id = session_id
    #     self.repo_node_name = node_name
    #     self.svs_cache_others = svs_cache_others
    #     self.svs = None
    #     self.global_view = global_view
    #     self.config = config
    #     self.expire_at = 0

    def heartbeat(self):
        # TODO: skip if expire_at value is big enough
        # hb tlv
        expire_at = int(time.time()+(self.config['period']*2))
        self.expire_at = expire_at
        favor = 1.85
        heartbeat_message_body = HeartbeatMessageBodyTlv()
        heartbeat_message_body.session_id = self.config['session_id'].encode()
        heartbeat_message_body.node_name = self.config['node_name'].encode()
        heartbeat_message_body.expire_at = expire_at
        heartbeat_message_body.favor = str(favor).encode()

        # hb msg
        heartbeat_message = MessageTlv()
        heartbeat_message.header = MessageTypes.HEARTBEAT
        heartbeat_message.body = heartbeat_message_body.encode()

        # heartbeat_message_body = HeartbeatMessageBody(self.svs_node_id, self.state_vector, heartbeat_message_body.encode())
        # print("state_vector: {0}".format(self.svs.getCore().getStateVector().to_str()))
        try:
            next_state_vector = self.svs.getCore().getStateTable().getSeqNum(Name.to_str(Name.from_str(self.config['session_id']))) + 1
        except TypeError:
            next_state_vector = 0
        self.global_view.update_session(self.config['session_id'], self.config['node_name'], expire_at, favor, next_state_vector)
        self.svs.publishData(heartbeat_message.encode())

    def detect_expired_sessions(self):
        deadline = int(time.time()) - (self.config['period'])
        expired_sessions = self.global_view.get_sessions_expired_by(deadline)
        for expired_session in expired_sessions:
            # generate expire msg and send
            # expire tlv
            expire_at = int(time.time()+(self.config['period']*2))
            favor = 1.85
            expire_message_body = ExpireMessageBodyTlv()
            expire_message_body.session_id = self.config['session_id'].encode()
            expire_message_body.node_name = self.config['node_name'].encode()
            expire_message_body.expire_at = expire_at
            expire_message_body.favor = str(favor).encode()
            expire_message_body.expired_session_id = expired_session['id'].encode()
            # expire msg
            expire_message = MessageTlv()
            expire_message.header = MessageTypes.EXPIRE
            expire_message.body = expire_message_body.encode()
            # apply globalview and send msg thru SVS
            self.global_view.expire_session(expired_session['id'])
            self.svs.publishData(expire_message.encode())
            val = "[MSG][EXPIRE]* sid={sid};exp_sid={esid}".format(
                sid=self.config['session_id'],
                esid=expired_session['id']
            )
            self.logger.info(val)

        # am I at the top of any insertion's backup list?
        underreplicated_insertions = self.global_view.get_underreplicated_insertions()
        for underreplicated_insertion in underreplicated_insertions:
            deficit = underreplicated_insertion['desired_copies'] - len(underreplicated_insertion['stored_bys'])
            for backuped_by in underreplicated_insertion['backuped_bys']:
                if (backuped_by['session_id'] == self.config['session_id']) and (backuped_by['rank'] < deficit):
                    self.fetch_file(underreplicated_insertion['id'], underreplicated_insertion['file_name'], underreplicated_insertion['packets'], underreplicated_insertion['digests'], underreplicated_insertion['fetch_path'])


    def claim(self):
        # TODO: possibility based on # active sessions and period
        # if random.random() < 0.618:
        #     return
        backupable_insertions = self.global_view.get_backupable_insertions()

        for backupable_insertion in backupable_insertions:

            if random.random() < 0.618:
                continue
            # print(json.dumps(backupable_insertion['stored_bys']))
            # print(json.dumps(backupable_insertion['backuped_bys']))
            already_in = False
            for stored_by in backupable_insertion['stored_bys']:
                if stored_by == self.config['session_id']:
                    already_in = True
                    break
            for backuped_by in backupable_insertion['backuped_bys']:
                if backuped_by['session_id'] == self.config['session_id']:
                    already_in = True
                    break
            if already_in == True:
                continue
            if len(backupable_insertion['backuped_bys']) == 0 and len(backupable_insertion['stored_bys']) == 0:
                continue
            authorizer = None
            if len(backupable_insertion['backuped_bys']) == 0:
                authorizer = {
                    'session_id': backupable_insertion['stored_bys'][-1],
                    'rank': -1,
                    'nonce': backupable_insertion['id']
                }
            else:
                authorizer = backupable_insertion['backuped_bys'][-1]
            # generate claim (request) msg and send
            # claim tlv
            expire_at = int(time.time()+(self.config['period']*2))
            favor = 1.85
            claim_message_body = ClaimMessageBodyTlv()
            claim_message_body.session_id = self.config['session_id'].encode()
            claim_message_body.node_name = self.config['node_name'].encode()
            claim_message_body.expire_at = expire_at
            claim_message_body.favor = str(favor).encode()
            claim_message_body.insertion_id = backupable_insertion['id'].encode()
            claim_message_body.type = ClaimMessageTypes.REQUEST
            claim_message_body.claimer_session_id = self.config['session_id'].encode()
            claim_message_body.claimer_nonce = secrets.token_hex(4).encode()
            claim_message_body.authorizer_session_id = authorizer['session_id'].encode()
            claim_message_body.authorizer_nonce = authorizer['nonce'].encode()
            # claim msg
            claim_message = MessageTlv()
            claim_message.header = MessageTypes.CLAIM
            claim_message.body = claim_message_body.encode()
            self.svs.publishData(claim_message.encode())
            val = "[MSG][CLAIM.R]*sid={sid};iid={iid}".format(
                sid=self.config['session_id'],
                iid=backupable_insertion['id']
            )
            self.logger.info(val)


    def periodic(self):
        # print('periodic')
        # periodic tasks:
        self.heartbeat()
        self.detect_expired_sessions()
        self.claim()
        # self.store()

        # sessions = self.global_view.get_sessions()
        # insertions = self.global_view.get_insertions()

        # for insertion in insertions:
        #     on = ""
        #     for stored_by in insertion['stored_bys']:
        #         on = on + stored_by + ","
        #     bck = ""
        #     for backuped_by in insertion['backuped_bys']:
        #         bck = bck + backuped_by['session_id'] + ","
        #     val = '[GV]           iid={iid}; name={name}; on={on}; bck={bck}'.format(
        #         iid=insertion['id'],
        #         name=insertion['file_name'],
        #         on=on,
        #         bck=bck
        #     )
            # self.logger.info(val)
        # print("--")

    def store(self, insertion_id: str):
        insertion = self.global_view.get_insertion(insertion_id)
        if len(insertion['stored_bys']) < insertion['desired_copies']:
            # store msg
            expire_at = int(time.time()+(self.config['period']*2))
            favor = 1.85
            store_message_body = StoreMessageBodyTlv()
            store_message_body.session_id = self.config['session_id'].encode()
            store_message_body.node_name = self.config['node_name'].encode()
            store_message_body.expire_at = expire_at
            store_message_body.favor = str(favor).encode()
            store_message_body.insertion_id = insertion_id.encode()
            # store msg
            store_message = MessageTlv()
            store_message.header = MessageTypes.STORE
            store_message.body = store_message_body.encode()
            # apply globalview and send msg thru SVS
            # next_state_vector = svs.getCore().getStateVector().get(config['session_id']) + 1

            self.global_view.store_file(insertion_id, self.config['session_id'])
            self.svs.publishData(store_message.encode())
            val = "[MSG][STORE]*  sid={sid};iid={iid}".format(
                sid=self.config['session_id'],
                iid=insertion_id
            )
            self.logger.info(val)

    def svs_missing_callback(self, missing_list):
        aio.ensure_future(self.on_missing_svs_messages(missing_list))

    async def on_missing_svs_messages(self, missing_list):
        for i in missing_list:
            while i.lowSeqNum <= i.highSeqNum:
                # print('{}:{}, {}'.format(i.nid, i.lowSeqNum, i.highSeqNum))
                message_bytes = await self.svs.fetchData(Name.from_str(i.nid), i.lowSeqNum)
                if message_bytes == None:
                    continue
                nid = i.nid
                seq = i.lowSeqNum
                message = Message(nid, seq, message_bytes)
                message_body = message.get_message_body()
                aio.ensure_future(message_body.apply(self.global_view, self.fetch_file, self.svs, self.config))
                # print('fetched GM {}:{}'.format(nid, seq))
                i.lowSeqNum = i.lowSeqNum + 1

    def svs_sending_callback(self, expire_at: int):
        self.expire_at = expire_at

    def fetch_file(self, insertion_id: str, file_name: str, packets: int, digests: List[bytes], fetch_path: str):
        val = "[ACT][FETCH]*  iid={iid};file_name={file_name};pcks={packets};fetch_path={fetch_path}".format(
            iid=insertion_id,
            file_name=file_name,
            packets=packets,
            fetch_path=fetch_path
        )
        self.logger.info(val)
        aio.ensure_future(self.async_fetch(insertion_id, file_name, packets, digests, fetch_path))

    async def async_fetch(self, insertion_id: str, file_name: str, packets: int, digests: List[bytes], fetch_path: str):
        self.logger.debug(packets)
        if packets > 1:
            start = time.time()
            inserted_packets = await self.fetch_segmented_file(file_name, packets, fetch_path)
            if inserted_packets == packets:
                end = time.time()
                duration = end -start
                val = "[ACT][FETCHED]*pcks={packets};duration={duration}".format(
                    packets=packets,
                    duration=duration
                )
                self.logger.info(val)
                self.store(insertion_id)
        elif packets == 1:
            inserted_packets = await self.fetch_single_file(file_name, fetch_path)
            if inserted_packets == packets:
                self.store(insertion_id)

    async def fetch_segmented_file(self, file_name: str, packets: int, fetch_path: str):
        semaphore = aio.Semaphore(10)
        fetched_segments = 0
        async for (_, _, content, data_bytes, key) in concurrent_fetcher(self.app, fetch_path, file_name, 0, packets-1, semaphore):
            #TODO: check digest
            # print("segment:")
            # print(Name.to_str(key))
            # print(type(content))
            # print(content)
            # print(type(data_bytes))
            # print(data_bytes)
            self.data_storage.put_data_packet(key, data_bytes)
            # self.data_storage.put_data_packet(key, content.tobytes())
            fetched_segments += 1
        return fetched_segments

    async def fetch_single_file(self, file_name: str, fetch_path: str):
        int_name = int_name = Name.normalize(fetch_path) + [Component.from_segment(0)]
        key = Name.normalize(file_name) + [Component.from_segment(0)]
        try:
            data_name, _, _, data_bytes = await self.app.express_interest(
                int_name, need_raw_packet=True, can_be_prefix=False, lifetime=1000)
        except InterestNack as e:
            return 0
        except InterestTimeout:
            return 0
        self.data_storage.put_data_packet(key, data_bytes)
        return 1


# ----------------------------------------------------------
# NDN Hydra Concurrent Fetcher
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import asyncio as aio
import logging
from ndn.app import NDNApp
from ndn.types import InterestNack, InterestTimeout
from ndn.encoding import Name, NonStrictName, Component
from typing import Optional

async def concurrent_fetcher(app: NDNApp, name: NonStrictName, start_block_id: int,
                             end_block_id: Optional[int], semaphore: aio.Semaphore, **kwargs):
    """
    An async-generator to fetch data packets between "`name`/`start_block_id`" and "`name`/`end_block_id`"\
        concurrently.
    :param app: NDNApp.
    :param name: NonStrictName. Name prefix of Data.
    :param start_block_id: int. The start segment number.
    :param end_block_id: Optional[int]. The end segment number. If not specified, continue fetching\
        until an interest receives timeout or nack or 3 times.
    :return: Yield ``(FormalName, MetaInfo, Content, RawPacket)`` tuples in order.
    """
    cur_id = start_block_id
    final_id = end_block_id if end_block_id is not None else 0x7fffffff
    is_failed = False
    tasks = []
    recv_window = cur_id - 1
    seq_to_data_packet = dict()    # Buffer for out-of-order delivery
    received_or_fail = aio.Event()

    async def _retry(seq: int):
        """
        Retry 3 times fetching data of the given sequence number or fail.
        :param seq: block_id of data
        """
        nonlocal app, name, semaphore, is_failed, received_or_fail, final_id
        int_name = name + [Component.from_segment(seq)]

        trial_times = 0
        while True:
            trial_times += 1
            if trial_times > 3:
                semaphore.release()
                is_failed = True
                received_or_fail.set()
                return
            try:
                logging.info('Express Interest: {}'.format(Name.to_str(int_name)))
                data_name, meta_info, content, data_bytes = await app.express_interest(
                    int_name, need_raw_packet=True, can_be_prefix=False, lifetime=1000, **kwargs)

                # Save data and update final_id
                logging.info('Received data: {}'.format(Name.to_str(data_name)))
                seq_to_data_packet[seq] = (data_name, meta_info, content, data_bytes)
                if meta_info is not None and meta_info.final_block_id is not None:
                    final_id = Component.to_number(meta_info.final_block_id)
                break
            except InterestNack as e:
                logging.info(f'Nacked with reason={e.reason}')
            except InterestTimeout:
                logging.info(f'Timeout')
        semaphore.release()
        received_or_fail.set()

    async def _dispatch_tasks():
        """
        Dispatch retry() tasks using semaphore.
        """
        nonlocal semaphore, tasks, cur_id, final_id, is_failed
        while cur_id <= final_id:
            await semaphore.acquire()
            if is_failed:
                received_or_fail.set()
                semaphore.release()
                break
            task = aio.get_event_loop().create_task(_retry(cur_id))
            tasks.append(task)
            cur_id += 1

    aio.get_event_loop().create_task(_dispatch_tasks())
    while 1:
        await received_or_fail.wait()
        received_or_fail.clear()
        # Re-assemble bytes in order
        while recv_window + 1 in seq_to_data_packet:
            yield seq_to_data_packet[recv_window + 1]
            del seq_to_data_packet[recv_window + 1]
            recv_window += 1
        # Return if all data have been fetched, or the fetching process failed
        if recv_window == final_id:
            await aio.gather(*tasks)
            return
        if is_failed:
            await aio.gather(*tasks)
            # New data may return during gather(), need to check again
            while recv_window + 1 in seq_to_data_packet:
                yield seq_to_data_packet[recv_window + 1]
                del seq_to_data_packet[recv_window + 1]
                recv_window += 1
            return


# ----------------------------------------------------------
# NDN Hydra Protocol Handle Base
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import logging
from ndn.app import NDNApp
from ndn.encoding import Name
from ndn_python_repo import Storage
from ndn_hydra.repo.utils.pubsub import PubSub

class ProtocolHandle(object):
    """
    Interface for protocol interest handles
    """
    def __init__(self, app: NDNApp, data_storage: Storage, pb: PubSub, config: dict):
        self.app = app
        self.data_storage = data_storage
        self.pb = pb
        self.config = config
        self.logger = logging.getLogger()
    async def listen(self, prefix: Name):
        raise NotImplementedError



# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

from .client import *
from .repo import *



# ----------------------------------------------------------
# NDN Hydra Repo Commands
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

from ndn.encoding import TlvModel, ModelField, NameField, UintField, RepeatedField, BytesField

class RepoTypeNumber:
  REPO_COMMAND = 201
  FILE = 202
  DESIRED_COPIES = 204
  PACKETS = 205
  DIGEST = 206
  SIZE = 207
  FETCH_PATH = 208
  SEQUENCE_NUMBER = 210
  # PUBLISHER_PREFIX = 211
  # NOTIFY_NONCE = 212

class FetchPath(TlvModel):
  prefix = NameField()

class File(TlvModel):
  file_name = NameField()
  desired_copies = UintField(RepoTypeNumber.DESIRED_COPIES)
  packets = UintField(RepoTypeNumber.PACKETS)
  digests = RepeatedField(BytesField(RepoTypeNumber.DIGEST))
  size = UintField(RepoTypeNumber.SIZE)

class FileList(TlvModel):
    list = RepeatedField(ModelField(RepoTypeNumber.FILE, File))

class RepoCommand(TlvModel):
  file = ModelField(RepoTypeNumber.FILE, File)
  sequence_number = UintField(RepoTypeNumber.SEQUENCE_NUMBER)
  fetch_path = ModelField(RepoTypeNumber.FETCH_PATH, FetchPath)


# ----------------------------------------------------------
# NDN Hydra Delete Command Handle
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import asyncio as aio
import logging
import time
from ndn.app import NDNApp
from ndn.encoding import Name, NonStrictName, Component
from ndn_python_repo import Storage
from ndn_hydra.repo.protocol.repo_commands import RepoCommand
from ndn_hydra.repo.utils.pubsub import PubSub
from ndn_hydra.repo.repo_messages.remove import RemoveMessageBodyTlv
from ndn_hydra.repo.repo_messages.message import MessageTlv, MessageTypes
from ndn_hydra.repo.main.main_loop import MainLoop
from ndn_hydra.repo.handle_protocol.protocol_handle_base import ProtocolHandle
from ndn_hydra.repo.global_view.global_view import GlobalView

class DeleteCommandHandle(ProtocolHandle):
    """
    DeleteCommandHandle processes delete command handles, and deletes corresponding data stored
    in the database.
    TODO: Add validator
    """
    def __init__(self, app: NDNApp, data_storage: Storage, pb: PubSub, config: dict,
                main_loop: MainLoop, global_view: GlobalView):
        """
        :param app: NDNApp.
        :param data_storage: Storage.
        :param pb: PubSub.
        :param config: All config Info.
        :param main_loop: SVS interface, Group Messages.
        :param global_view: Global View.
        """
        super(DeleteCommandHandle, self).__init__(app, data_storage, pb, config)
        self.prefix = None
        self.main_loop = main_loop
        self.global_view = global_view
        self.repo_prefix = config['repo_prefix']
        #self.register_root = config['repo_config']['register_root']

    async def listen(self, prefix: NonStrictName):
        """
        Register routes for command interests.
        This function needs to be called explicitly after initialization.
        :param name: NonStrictName. The name prefix to listen on.
        """
        self.prefix = prefix
        self.logger.info(f'Insert handle: subscribing to {Name.to_str(self.prefix) + "/delete"}')
        self.pb.subscribe(self.prefix + ['delete'], self._on_delete_msg)
        # start to announce process status
        # await self._schedule_announce_process_status(period=3)

    def _on_delete_msg(self, msg):

        try:
            cmd = RepoCommand.parse(msg)
            # if cmd.name == None:
            #     raise DecodeError()
        except (DecodeError, IndexError) as exc:
            logging.warning('Parameter interest decoding failed')
            return
        aio.ensure_future(self._process_delete(cmd))

    async def _process_delete(self, cmd: RepoCommand):
        """
        Process delete command.
        """

        file_name = cmd.file.file_name
        sequence_number = cmd.sequence_number

        self.logger.info("[cmd][DELETE] file {}".format(Name.to_str(file_name)))

        insertion = self.global_view.get_insertion_by_file_name(Name.to_str(file_name))
        if insertion == None:
            self.logger.warning("file does not exist")
            return

        insertion_id = insertion['id']
        # add tlv
        expire_at = int(time.time()+(self.config['period']*2))
        favor = 1.85
        remove_message_body = RemoveMessageBodyTlv()
        remove_message_body.session_id = self.config['session_id'].encode()
        remove_message_body.node_name = self.config['node_name'].encode()
        remove_message_body.expire_at = expire_at
        remove_message_body.favor = str(favor).encode()
        remove_message_body.insertion_id = insertion_id.encode()
        # add msg
        remove_message = MessageTlv()
        remove_message.header = MessageTypes.REMOVE
        remove_message.body = remove_message_body.encode()
        self.global_view.delete_insertion(insertion_id)
        self.main_loop.svs.publishData(remove_message.encode())
        val = "[MSG][REMOVE]* iid={iid}".format(
            iid=insertion_id
        )
        self.logger.info(val)









# ----------------------------------------------------------
# NDN Hydra Delete Client
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import logging
from ndn.app import NDNApp
from ndn.encoding import Name, Component, FormalName
from ndn_hydra.repo.protocol.repo_commands import RepoCommand, File, FetchPath
from ndn_hydra.repo.utils.pubsub import PubSub

class HydraDeleteClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
      """
      This client deletes a certain file from the remote repo.
      :param app: NDNApp.
      :param client_prefix: NonStrictName. Routable name to client.
      :param repo_prefix: NonStrictName. Routable name to remote repo.
      """
      self.app = app
      self.client_prefix = client_prefix
      self.repo_prefix = repo_prefix
      self.pb = PubSub(self.app, self.client_prefix)

    async def delete_file(self, file_name: FormalName) -> bool:
      """
      Delete a file asscoiated with a file name from the remote repo
      """
      # send command interest
      file = File()
      file.file_name = file_name
      file.desired_copies = 0
      file.packets = 0
      file.digests = []
      file.size = 0
      fetch_path = FetchPath()
      fetch_path.prefix = self.client_prefix + [Component.from_str("upload")] + file_name
      cmd = RepoCommand()
      cmd.file = file
      cmd.sequence_number = 0
      cmd.fetch_path = fetch_path
      cmd_bytes = cmd.encode()

      # publish msg to repo's delete topic
      await self.pb.wait_for_ready()
      print(Name.to_str(self.repo_prefix + ['delete']))
      is_success = await self.pb.publish(self.repo_prefix + ['delete'], cmd_bytes)
      if is_success:
          logging.info('Published an delete msg and was acknowledged by a subscriber')
      else:
          logging.info('Published an delete msg but was not acknowledged by a subscriber')
      return is_success


# ----------------------------------------------------------
# NDN Hydra Insert Command Handle
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import asyncio as aio
import logging
import random
import secrets
import time
from ndn.app import NDNApp
from ndn.encoding import Name, NonStrictName, Component, DecodeError
from ndn_python_repo import Storage
from ndn_hydra.repo.handle_protocol.protocol_handle_base import ProtocolHandle
from ndn_hydra.repo.protocol.repo_commands import RepoCommand
from ndn_hydra.repo.utils.pubsub import PubSub
from ndn_hydra.repo.global_view.global_view import GlobalView
from ndn_hydra.repo.repo_messages.add import FileTlv, FetchPathTlv, BackupTlv, AddMessageBodyTlv
from ndn_hydra.repo.repo_messages.message import MessageTlv, MessageTypes
from ndn_hydra.repo.main.main_loop import MainLoop

class InsertCommandHandle(ProtocolHandle):
    """
    InsertCommandHandle processes insert command handles, and deletes corresponding data stored
    in the database.
    """
    def __init__(self, app: NDNApp, data_storage: Storage, pb: PubSub, config: dict,
                main_loop: MainLoop, global_view: GlobalView):
        """
        :param app: NDNApp.
        :param data_storage: Storage.
        :param pb: PubSub.
        :param config: All config Info.
        :param main_loop: the Main Loop.
        :param global_view: Global View.
        """
        super(InsertCommandHandle, self).__init__(app, data_storage, pb, config)
        self.prefix = None
        self.main_loop = main_loop
        self.global_view = global_view
        self.repo_prefix = config['repo_prefix']

    async def listen(self, prefix: NonStrictName):
        """
        Register routes for command interests.
        This function needs to be called explicitly after initialization.
        :param name: NonStrictName. The name prefix to listen on.
        """
        self.prefix = prefix
        self.logger.info(f'Insert handle: subscribing to {Name.to_str(self.prefix) + "/insert"}')
        self.pb.subscribe(self.prefix + ['insert'], self._on_insert_msg)
        # start to announce process status
        # await self._schedule_announce_process_status(period=3)

    def _on_insert_msg(self, msg):
        try:
            cmd = RepoCommand.parse(msg)
            #if cmd.file == None:
            #    raise DecodeError()
        except (DecodeError, IndexError) as exc:
            logging.warning('Parameter interest decoding failed')
            return
        aio.ensure_future(self._process_insert(cmd))

    async def _process_insert(self, cmd: RepoCommand):
        """
        Process insert command.
        """
        # print("Process Insert Command for File: ")
        # print("receive INSERT command for file: {}".format(Name.to_str(cmd.file.file_name)))

        file_name = cmd.file.file_name
        desired_copies = cmd.file.desired_copies
        packets = cmd.file.packets
        digests = cmd.file.digests
        size = cmd.file.size
        sequence_number = cmd.sequence_number
        fetch_path = cmd.fetch_path.prefix

        self.logger.info("[cmd][INSERT] file {}".format(Name.to_str(file_name)))

        # TODO: check duplicate sequence number

        # are there enough sessions?
        if desired_copies == 0:
            self.logger.warning("desired_copies is 0")
            return

        sessions = self.global_view.get_sessions()

        if len(sessions) < (desired_copies * 2):
            self.logger.warning("not enough node sessions") # TODO: notify the client?
            return

        # generate unique insertion_id
        insertion_id = secrets.token_hex(8)
        while self.global_view.get_insertion(insertion_id) != None:
            insertion_id = secrets.token_hex(8)

        # select sessions
        random.shuffle(sessions)
        picked_sessions = random.sample(sessions, (desired_copies * 2))

        pickself = False
        for i in range(desired_copies):
            if picked_sessions[i]['id'] == self.config['session_id']:
                pickself = True
                break

        # if pickself:
        #     # TODO: fetch and store this file
        #     pass
            # print("pick myself")
            # picked_sessions = list(filter(lambda x: x['id'] != self.config['session_id'], picked_sessions))


        backups = []
        backup_list = []
        for picked_session in picked_sessions:
            session_id = picked_session['id']
            nonce = secrets.token_hex(4)
            backup = BackupTlv()
            backup.session_id = session_id.encode()
            backup.nonce = nonce.encode()
            backups.append(backup)
            backup_list.append((session_id, nonce))


        # add tlv
        expire_at = int(time.time()+(self.config['period']*2))
        favor = 1.85
        add_message_body = AddMessageBodyTlv()
        add_message_body.session_id = self.config['session_id'].encode()
        add_message_body.node_name = self.config['node_name'].encode()
        add_message_body.expire_at = expire_at
        add_message_body.favor = str(favor).encode()
        add_message_body.insertion_id = insertion_id.encode()
        add_message_body.file = FileTlv()
        add_message_body.file.file_name = file_name
        add_message_body.file.desired_copies = desired_copies
        add_message_body.file.packets = packets
        add_message_body.file.digests = digests
        add_message_body.file.size = size
        add_message_body.sequence_number = sequence_number
        add_message_body.fetch_path = FetchPathTlv()
        add_message_body.fetch_path.prefix = fetch_path
        # add_message_body.is_stored_by_origin = 1 if pickself else 0
        add_message_body.is_stored_by_origin = 0
        add_message_body.backup_list = backups
        # add msg
        add_message = MessageTlv()
        add_message.header = MessageTypes.ADD
        add_message.body = add_message_body.encode()
        # apply globalview and send msg thru SVS
        try:
            next_state_vector = self.main_loop.svs.getCore().getStateTable().getSeqNum(Name.to_str(Name.from_str(self.config['session_id']))) + 1
        except TypeError:
            next_state_vector = 0
        self.global_view.add_insertion(
            insertion_id,
            Name.to_str(file_name),
            sequence_number,
            size,
            self.config['session_id'],
            Name.to_str(fetch_path),
            next_state_vector,
            b''.join(digests),
            packets=packets,
            desired_copies=desired_copies
        )
        if pickself:
            # self.global_view.store_file(insertion_id, self.config['session_id'])
            self.main_loop.fetch_file(insertion_id, Name.to_str(file_name), packets, digests, Name.to_str(fetch_path))
        self.global_view.set_backups(insertion_id, backup_list)
        self.main_loop.svs.publishData(add_message.encode())
        bak = ""
        for backup in backup_list:
            bak = bak + backup[0] + ","
        val = "[MSG][ADD]*    sid={sid};iid={iid};file={fil};cop={cop};pck={pck};siz={siz};seq={seq};slf={slf};bak={bak}".format(
            sid=self.config['session_id'],
            iid=insertion_id,
            fil=Name.to_str(file_name),
            cop=desired_copies,
            pck=packets,
            siz=size,
            seq=sequence_number,
            # slf=1 if pickself else 0,
            slf=0,
            bak=bak
        )
        self.logger.info(val)


# ----------------------------------------------------------
# NDN Hydra Expire Group Message
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

from typing import Callable
from ndn.encoding import *
import time
from ndn_hydra.repo.repo_messages.message_base import MessageBodyBase
from ndn_hydra.repo.repo_messages.store import StoreMessageBodyTlv
from ndn_hydra.repo.global_view.global_view import GlobalView

class ExpireMessageBodyTypes:
    SESSION_ID = 83
    NODE_NAME = 84
    EXPIRE_AT = 85
    FAVOR = 86

    EXPIRED_SESSION_ID = 90

class ExpireMessageBodyTlv(TlvModel):
    session_id = BytesField(ExpireMessageBodyTypes.SESSION_ID)
    node_name = BytesField(ExpireMessageBodyTypes.NODE_NAME)
    expire_at = UintField(ExpireMessageBodyTypes.EXPIRE_AT)
    favor = BytesField(ExpireMessageBodyTypes.FAVOR)
    expired_session_id = BytesField(ExpireMessageBodyTypes.EXPIRED_SESSION_ID)

class ExpireMessageBody(MessageBodyBase):
    def __init__(self, nid:str, seq:int, raw_bytes:bytes):
        super(ExpireMessageBody, self).__init__(nid, seq)
        self.message_body = ExpireMessageBodyTlv.parse(raw_bytes)

    async def apply(self, global_view: GlobalView, fetch_file: Callable, svs, config):
        session_id = self.message_body.session_id.tobytes().decode()
        node_name = self.message_body.node_name.tobytes().decode()
        expire_at = self.message_body.expire_at
        favor = float(self.message_body.favor.tobytes().decode())
        expired_session_id = self.message_body.expired_session_id.tobytes().decode()
        val = "[MSG][EXPIRE]  sid={sid};exp_sid={esid}".format(
            sid=session_id,
            esid=expired_session_id
        )
        self.logger.info(val)
        global_view.expire_session(expired_session_id)
        # am I at the top of any insertion's backup list?
        underreplicated_insertions = global_view.get_underreplicated_insertions()
        from .message import MessageTlv, MessageTypes
        for underreplicated_insertion in underreplicated_insertions:
            deficit = underreplicated_insertion['desired_copies'] - len(underreplicated_insertion['stored_bys'])
            for backuped_by in underreplicated_insertion['backuped_bys']:
                if (backuped_by['session_id'] == config['session_id']) and (backuped_by['rank'] < deficit):

                    digests = underreplicated_insertion['digests']
                    self.logger.debug(type(digests[0]))

                    fetch_file(underreplicated_insertion['id'], underreplicated_insertion['file_name'], underreplicated_insertion['packets'], underreplicated_insertion['digests'], underreplicated_insertion['fetch_path'])

                    # # generate store msg and send
                    # # store tlv
                    # expire_at = int(time.time()+(config['period']*2))
                    # favor = 1.85
                    # store_message_body = StoreMessageBodyTlv()
                    # store_message_body.session_id = config['session_id'].encode()
                    # store_message_body.node_name = config['node_name'].encode()
                    # store_message_body.expire_at = expire_at
                    # store_message_body.favor = str(favor).encode()
                    # store_message_body.insertion_id = underreplicated_insertion['id'].encode()
                    # # store msg
                    # store_message = MessageTlv()
                    # store_message.header = MessageTypes.STORE
                    # store_message.body = store_message_body.encode()
                    # # apply globalview and send msg thru SVS
                    # # next_state_vector = svs.getCore().getStateVector().get(config['session_id']) + 1
                    # global_view.store_file(underreplicated_insertion['id'], config['session_id'])
                    # svs.publishData(store_message.encode())
                    # val = "[MSG][STORE]+  sid={sid};iid={iid}".format(
                    #     sid=config['session_id'],
                    #     iid=underreplicated_insertion['id']
                    # )
                    # self.logger.info(val)
        # update session
        global_view.update_session(session_id, node_name, expire_at, favor, self.seq)
        return


# ----------------------------------------------------------
# NDN Hydra Insert Client
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import logging
from hashlib import blake2b
from ndn.app import NDNApp
from ndn.encoding import Name, Component, FormalName
from ndn_hydra.repo.protocol.repo_commands import RepoCommand, File, FetchPath
from ndn_hydra.repo.utils.pubsub import PubSub

SEGMENT_SIZE = 8192

class HydraInsertClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
      """
      This client inserts data packets from the remote repo.
      :param app: NDNApp.
      :param client_prefix: NonStrictName. Routable name to client.
      :param repo_prefix: NonStrictName. Routable name to remote repo.
      """
      self.app = app
      self.client_prefix = client_prefix
      self.repo_prefix = repo_prefix
      self.pb = PubSub(self.app, self.client_prefix)
      self.packets = []
      self.digests = []

    async def insert_file(self, file_name: FormalName, desired_copies: int, path: str) -> bool:
      """
      Insert a file associated with a file name to the remote repo
      """
      # send command interest

      test_name = file_name + [Component.from_version(3)] + [Component.from_segment(1)]
      print(Name.to_str(test_name))

      size = 0
      fetch_file_prefix = self.client_prefix + [Component.from_str("upload")] + file_name

      with open(path, "rb") as f:
        data = f.read()
        size = len(data)
        print("size: {0}".format(size))
        seg_cnt = (len(data) + SEGMENT_SIZE - 1) // SEGMENT_SIZE
        packets = seg_cnt
        self.packets = [self.app.prepare_data(fetch_file_prefix + [Component.from_segment(i)],
                                              data[i*SEGMENT_SIZE:(i+1)*SEGMENT_SIZE],
                                              freshness_period=10000,
                                              final_block_id=Component.from_segment(seg_cnt - 1))
                        for i in range(seg_cnt)]

        self.digests = [bytes(blake2b(data[i*SEGMENT_SIZE:(i+1)*SEGMENT_SIZE]).digest()[:2]) for i in range(seg_cnt)]
        print(self.digests[0].hex())

      print(f'Created {seg_cnt} chunks under name {Name.to_str(fetch_file_prefix)}')

      @self.app.route(fetch_file_prefix)
      def on_interest(int_name, _int_param, _app_param):
        if Component.get_type(int_name[-1]) == Component.TYPE_SEGMENT:
            seg_no = Component.to_number(int_name[-1])
        else:
            seg_no = 0
        if seg_no < seg_cnt:
            self.app.put_raw_packet(self.packets[seg_no])

      file = File()
      file.file_name = file_name
      file.desired_copies = desired_copies
      file.packets = packets
      file.digests = self.digests
      file.size = size
      fetch_path = FetchPath()
      fetch_path.prefix = fetch_file_prefix
      cmd = RepoCommand()
      cmd.file = file
      cmd.sequence_number = 0
      cmd.fetch_path = fetch_path
      cmd_bytes = cmd.encode()

      # publish msg to repo's insert topic
      await self.pb.wait_for_ready()
      print(Name.to_str(self.repo_prefix + ['insert']))
      is_success = await self.pb.publish(self.repo_prefix + ['insert'], cmd_bytes)
      if is_success:
          logging.info('Published an insert msg and was acknowledged by a subscriber')
      else:
          logging.info('Published an insert msg but was not acknowledged by a subscriber')
      return is_success


# ----------------------------------------------------------
# NDN Hydra Read Handle
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import asyncio as aio
import logging
from secrets import choice
from ndn.app import NDNApp
from ndn.encoding import Name, ContentType, Component, parse_data
from ndn_python_repo import Storage
from ndn_hydra.repo.global_view.global_view import GlobalView

class ReadHandle(object):
    """
    ReadHandle processes ordinary interests, and return corresponding data if exists.
    """
    def __init__(self, app: NDNApp, data_storage: Storage, global_view: GlobalView, config: dict):
        """
        :param app: NDNApp.
        :param data_storage: Storage.
        :param global_view: Global View.
        :param config: All config Info.
        """
        self.app = app
        self.data_storage = data_storage
        self.global_view = global_view
        self.session_id = config['session_id']
        self.repo_prefix = config['repo_prefix']

        self.logger = logging.getLogger()

        self.command_comp = "/fetch"
        self.sid_comp = "/sid"

        self.listen(Name.from_str(self.repo_prefix + self.command_comp))
        self.listen(Name.from_str(self.repo_prefix + self.sid_comp  + "/" + self.session_id + self.command_comp))

    def listen(self, prefix):
        """
        This function needs to be called for prefix of all data stored.
        :param prefix: NonStrictName.
        """
        self.app.route(prefix)(self._on_interest)
        self.logger.info(f'Read handle: listening to {Name.to_str(prefix)}')

    def unlisten(self, prefix):
        """
        :param name: NonStrictName.
        """
        aio.ensure_future(self.app.unregister(prefix))
        self.logger.info(f'Read handle: stop listening to {Name.to_str(prefix)}')

    def _on_interest(self, int_name, int_param, _app_param):
        """
        Repo should not respond to any interest with MustBeFresh flag set.
        Repo will:
        - Reply with data of its own
        - Nack if data can not be found within the repo
        - Reply with a redirect to another node
        Assumptions:
        - A node on the on list will have the file in complete form
        """
        if int_param.must_be_fresh:
            return
        # get rid of the security part if any on the int_name
        file_name = self._get_file_name_from_interest(Name.to_str(int_name[:-1]))
        best_id = self._best_id_for_file(file_name)
        segment_comp = "/" + Component.to_str(int_name[-1])

        if best_id == self.session_id:
            if segment_comp == "/seg=0":
                self.logger.info(f'[cmd][FETCH] serving data to client')

            # serving my own data
            data_bytes = self.data_storage.get_data_packet(file_name + segment_comp, int_param.can_be_prefix)
            if data_bytes == None:
                return

            self.logger.info(f'Read handle: serve data {Name.to_str(int_name)}')
            _, _, content, _ = parse_data(data_bytes)
            # print("serve"+file_name + segment_comp+"   "+Name.to_str(name))
            final_id = Component.from_number(int(self.global_view.get_insertion_by_file_name(file_name)["packets"])-1, Component.TYPE_SEGMENT)
            self.app.put_data(int_name, content=content, content_type=ContentType.BLOB, final_block_id=final_id)
            return

        elif best_id == None:
            if segment_comp == "/seg=0":
                self.logger.info(f'[cmd][FETCH] nacked client due to no file in repo')

            # nack due to lack of avaliability
            self.app.put_data(int_name, content=None, content_type=ContentType.NACK)
            self.logger.info(f'Read handle: data not found {Name.to_str(int_name)}')
            return
        else:
            if segment_comp == "/seg=0":
                self.logger.info(f'[cmd][FETCH] linked to another node in the repo')

            # create a link to a node who has the content
            new_name = self.repo_prefix + self.sid_comp + "/" + best_id + self.command_comp + file_name
            link_content = bytes(new_name.encode())
            final_id = Component.from_number(int(self.global_view.get_insertion_by_file_name(file_name)["packets"])-1, Component.TYPE_SEGMENT)
            self.app.put_data(int_name, content=link_content, content_type=ContentType.LINK, final_block_id=final_id)
            self.logger.info(f'Read handle: redirected {Name.to_str(int_name)}')
            return

    def _get_file_name_from_interest(self, int_name):
        file_name = int_name[len(self.repo_prefix):]
        if file_name[0:len(self.sid_comp)] == self.sid_comp:
            return file_name[(len(self.sid_comp)+len("/" + self.session_id)+len(self.command_comp)):]
        else:
            return file_name[(len(self.command_comp)):]

    def _best_id_for_file(self, file_name: str):
        file_info = self.global_view.get_insertion_by_file_name(file_name)
        if file_info != None:
            on_list = file_info["stored_bys"]
            if file_info["is_deleted"] == True or not on_list:
                return None
            if self.session_id in on_list:
                return self.session_id
            else:
              return choice(on_list)
        return None


# ----------------------------------------------------------
# NDN Hydra Message Body Base
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import logging

class MessageBodyBase:
    def __init__(self, nid:str, seq:int):
        self.nid = nid
        self.seq = seq
        self.logger = logging.getLogger()

    async def apply(self, global_view):
        raise NotImplementedError


# ----------------------------------------------------------
# NDN Hydra Global View
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import os
import sqlite3
from sqlite3 import Error
from typing import List, Tuple

sql_create_sessions_tables = """
CREATE TABLE IF NOT EXISTS sessions (
    id TEXT PRIMARY KEY,
    node_name TEXT NOT NULL,
    expire_at INTEGER NOT NULL,
    favor REAL NOT NULL,
    state_vector INTEGER NOT NULL,
    is_expired INTEGER NOT NULL DEFAULT 0
);
"""
sql_create_insertions_tables = """
CREATE TABLE IF NOT EXISTS insertions (
    id TEXT PRIMARY KEY,
    file_name TEXT NOT NULL,
    sequence_number INTEGER NOT NULL,
    desired_copies INTEGER NOT NULL DEFAULT 3,
    packets INTEGER NOT NULL DEFAULT 1,
    digests BLOB NOT NULL DEFAULT 0,
    size INTEGER NOT NULL,
    origin_session_id TEXT NOT NULL,
    fetch_path TEXT NOT NULL,
    state_vector INTEGER NOT NULL,
    is_deleted INTEGER NOT NULL DEFAULT 0
);
"""
sql_create_stored_by_tables = """
CREATE TABLE IF NOT EXISTS stored_by (
    id INTEGER PRIMARY KEY,
    insertion_id TEXT NOT NULL,
    session_id NOT NULL
);
"""
sql_create_backuped_by_tables = """
CREATE TABLE IF NOT EXISTS backuped_by (
    id INTEGER PRIMARY KEY,
    insertion_id TEXT NOT NULL,
    session_id NOT NULL,
    rank INTEGER NOT NULL,
    nonce TEXT NOT NULL
);
"""
sql_create_pending_stores_tables = """
CREATE TABLE IF NOT EXISTS pending_stores (
    id INTEGER PRIMARY KEY,
    insertion_id TEXT NOT NULL,
    session_id NOT NULL
);
"""

class GlobalView:

    def __init__(self, db: str):
        self.db = os.path.expanduser(db)
        if len(os.path.dirname(self.db)) > 0 and not os.path.exists(os.path.dirname(self.db)):
            try:
                os.makedirs(os.path.dirname(self.db))
            except PermissionError:
                raise PermissionError(f'Could not create database directory: {self.db}') from None
            except FileExistsError:
                pass
        self.__create_tables()

    def __get_connection(self):
        conn = None
        try:
            conn = sqlite3.connect(self.db)
        except Error as e:
            print(e)
        return conn

    def __execute_sql(self, sql: str):
        # print(sql)
        result = []
        conn = self.__get_connection()
        if conn is not None:
            try:
                c = conn.cursor()
                c.execute(sql)
                conn.commit()
                result = c.fetchall()
            except Error as e:
                print(e)
            conn.close()
        return result

    def __execute_sql_qmark(self, sql: str, par: Tuple):
        # print(sql)
        result = []
        conn = self.__get_connection()
        if conn is not None:
            try:
                c = conn.cursor()
                c.execute(sql, par)
                conn.commit()
                result = c.fetchall()
            except Error as e:
                print(e)
            conn.close()
        return result

    def __create_tables(self):
        self.__execute_sql(sql_create_sessions_tables)
        self.__execute_sql(sql_create_insertions_tables)
        self.__execute_sql(sql_create_stored_by_tables)
        self.__execute_sql(sql_create_backuped_by_tables)
        self.__execute_sql(sql_create_pending_stores_tables)

    def __rerank_backuped_by(self, insertion_id: str, session_id: str):
        # check rank
        sql_get_backup = """
        SELECT DISTINCT insertion_id, session_id, rank
        FROM backuped_by
        WHERE (insertion_id = ?) AND (session_id = ?)
        """
        result = self.__execute_sql_qmark(sql_get_backup, (insertion_id, session_id))
        # return result
        rank = -1
        if len(result) == 1:
            rank = result[0][2]
        if rank != -1:
            sql_rerank = """
            UPDATE backuped_by
            SET rank = rank - 1
            WHERE (insertion_id = ?) AND (rank > ?)
            """
            self.__execute_sql_qmark(sql_rerank, (insertion_id, rank))

    def get_session(self, session_id: str):
        sql = """
        SELECT DISTINCT
            id, node_name, expire_at, favor, state_vector, is_expired
        FROM
            sessions
        WHERE
            id = ?
        """
        result = self.__execute_sql_qmark(sql, (session_id, ))
        if len(result) != 1:
            return None
        else:
            return {
                'id': result[0][0],
                'node_name': result[0][1],
                'expire_at': result[0][2],
                'favor': result[0][3],
                'state_vector': result[0][4],
                'is_expired': False if (result[0][5] == 0) else True
            }

    def get_sessions(self, including_expired: bool = False):
        if including_expired:
            sql = """
            SELECT DISTINCT
                id, node_name, expire_at, favor, state_vector, is_expired
            FROM
                sessions
            """
        else:
            sql = """
            SELECT DISTINCT
                id, node_name, expire_at, favor, state_vector, is_expired
            FROM
                sessions
            WHERE
                is_expired = 0
            """
        results = self.__execute_sql(sql)
        sessions = []
        for result in results:
            sessions.append({
                'id': result[0],
                'node_name': result[1],
                'expire_at': result[2],
                'favor': result[3],
                'state_vector': result[4],
                'is_expired': False if (result[5] == 0) else True
            })
        return sessions

    def get_sessions_expired_by(self, timestamp: int):
        sql = """
        SELECT DISTINCT
            id, node_name, expire_at, favor, state_vector, is_expired
        FROM
            sessions
        WHERE
            is_expired = 0 AND
            expire_at <= ?
        """
        results = self.__execute_sql_qmark(sql, (timestamp, ))
        sessions = []
        for result in results:
            sessions.append({
                'id': result[0],
                'node_name': result[1],
                'expire_at': result[2],
                'favor': result[3],
                'state_vector': result[4],
                'is_expired': False if (result[5] == 0) else True
            })
        return sessions

    def __add_session(self, session_id: str, node_name: str, expire_at: int, favor: float, state_vector: int):
        # start session
        sql = """
        INSERT OR IGNORE INTO sessions
            (id, node_name, expire_at, favor, state_vector, is_expired)
        VALUES
            (?, ?, ?, ?, ?, 0)
        """
        self.__execute_sql_qmark(sql, (session_id, node_name, expire_at, favor, state_vector))

    def update_session(self, session_id: str, node_name: str, expire_at: int, favor: float, state_vector: int):
        self.__add_session(session_id, node_name, expire_at, favor, state_vector)
        sql = """
        UPDATE sessions
        SET expire_at = ?,
            favor = ?,
            state_vector = ?
        WHERE
            id = ?
        """
        self.__execute_sql_qmark(sql, (expire_at, favor, state_vector, session_id))

    def expire_session(self, session_id: str):

        # stored_by
        sql_stored_by = """
        DELETE FROM stored_by WHERE session_id = ?
        """
        self.__execute_sql_qmark(sql_stored_by, (session_id, ))

        # backuped_by
        sql_get_backups = """
        SELECT insertion_id, session_id
        FROM backuped_by
        WHERE session_id = ?
        """
        backups = self.__execute_sql_qmark(sql_get_backups, (session_id, ))
        for backup in backups:
            # rerank
            self.__rerank_backuped_by(backup[0], session_id)
        # remove
        sql_delete_backuped_by = """
        DELETE FROM backuped_by WHERE session_id = ?
        """
        self.__execute_sql_qmark(sql_delete_backuped_by, (session_id, ))

        # pending_stores
        sql_pending_stores = """
        DELETE FROM pending_stores WHERE session_id = ?
        """
        self.__execute_sql_qmark(sql_pending_stores, (session_id, ))

        # expire session
        sql = """
        UPDATE sessions
        SET is_expired = 1
        WHERE id = ?
        """
        self.__execute_sql_qmark(sql, (session_id, ))

    def __split_digests(self, digests: bytes, size: int):
        digests_bytes = bytes(digests)
        return [digests_bytes[i:i+size] for i in range(0, len(digests_bytes), size)]


    def get_insertion(self, insertion_id: str):
        sql = """
        SELECT DISTINCT
            id, file_name, sequence_number, desired_copies, packets, size, origin_session_id, fetch_path, state_vector, is_deleted, digests
        FROM
            insertions
        WHERE
            id = ?
        """
        result = self.__execute_sql_qmark(sql, (insertion_id, ))
        if len(result) != 1:
            return None
        else:
            return {
                'id': result[0][0],
                'file_name': result[0][1],
                'sequence_number': result[0][2],
                'desired_copies': result[0][3],
                'packets': result[0][4],
                'size': result[0][5],
                'origin_session_id': result[0][6],
                'fetch_path': result[0][7],
                'state_vector': result[0][8],
                'is_deleted': False if (result[0][9] == 0) else True,
                'digests': self.__split_digests(result[0][10], 2),
                'stored_bys': self.get_stored_bys(result[0][0]),
                'backuped_bys': self.get_backuped_bys(result[0][0])
            }

    def get_insertions(self, including_deleted: bool = False):
        if including_deleted:
            sql = """
            SELECT DISTINCT
                id, file_name, sequence_number, desired_copies, packets, size, origin_session_id, fetch_path, state_vector, is_deleted, digests
            FROM
                insertions
            """
        else:
            sql = """
            SELECT DISTINCT
                id, file_name, sequence_number, desired_copies, packets, size, origin_session_id, fetch_path, state_vector, is_deleted, digests
            FROM
                insertions
            WHERE
                is_deleted = 0
            """
        results = self.__execute_sql(sql)
        insertions = []
        for result in results:
            insertions.append({
                'id': result[0],
                'file_name': result[1],
                'sequence_number': result[2],
                'desired_copies': result[3],
                'packets': result[4],
                'size': result[5],
                'origin_session_id': result[6],
                'fetch_path': result[7],
                'state_vector': result[8],
                'is_deleted': False if (result[9] == 0) else True,
                'digests': self.__split_digests(result[10], 2),
                'stored_bys': self.get_stored_bys(result[0]),
                'backuped_bys': self.get_backuped_bys(result[0])
            })
        return insertions

    def get_insertion_by_file_name(self, file_name: str):
        sql = """
        SELECT DISTINCT
            id, file_name, sequence_number, desired_copies, packets, size, origin_session_id, fetch_path, state_vector, is_deleted, digests
        FROM
            insertions
        WHERE
            file_name = ? AND
            is_deleted = 0
        """
        result = self.__execute_sql_qmark(sql, (file_name, ))
        if len(result) != 1:
            return None
        else:
            return {
                'id': result[0][0],
                'file_name': result[0][1],
                'sequence_number': result[0][2],
                'desired_copies': result[0][3],
                'packets': result[0][4],
                'size': result[0][5],
                'origin_session_id': result[0][6],
                'fetch_path': result[0][7],
                'state_vector': result[0][8],
                'is_deleted': False if (result[0][9] == 0) else True,
                'digests': self.__split_digests(result[0][10], 2),
                'stored_bys': self.get_stored_bys(result[0][0]),
                'backuped_bys': self.get_backuped_bys(result[0][0])
            }

    def get_underreplicated_insertions(self):
        insertions = self.get_insertions()
        underreplicated_insertions = []
        for insertion in insertions:
            if len(insertion['stored_bys']) < insertion['desired_copies']:
                underreplicated_insertions.append(insertion)
        return underreplicated_insertions

    def get_backupable_insertions(self):
        insertions = self.get_insertions()
        backupable_insertions = []
        for insertion in insertions:
            # print("stored_bys: {}; backuped_bys: {}; dcopies: {}".format(len(insertion['stored_bys']), len(insertion['backuped_bys']), insertion['desired_copies']))
            if ( len(insertion['stored_bys']) + len(insertion['backuped_bys']) ) < (insertion['desired_copies'] * 2):
                backupable_insertions.append(insertion)
        return backupable_insertions

    def add_insertion(self, insertion_id: str, file_name: str, sequence_number: int, size: int, origin_session_id: str,
               fetch_path: str, state_vector: int, digests: bytes, packets=1, desired_copies=3):
        # # check (same insertion_id):
        # insertion = self.get_insertion(insertion_id)
        # if (insertion is not None) and (insertion['is_deleted'] == False):
        #     return False
        # # TODO(check if there are insertions with the same file_name)
        # # TODO(may not needed, check if there are insertions with the same file_name and same sequence_number)

        sql = """
        INSERT OR IGNORE INTO insertions
            (id, file_name, sequence_number, desired_copies, packets, size, origin_session_id, fetch_path, state_vector, is_deleted, digests)
        VALUES
            (?, ?, ?, ?, ?, ?, ?, ?, ?, 0, ?)
        """
        self.__execute_sql_qmark(sql, (insertion_id, file_name, sequence_number, desired_copies, packets, size, origin_session_id, fetch_path, state_vector, digests))

    def delete_insertion(self, insertion_id: str):
        # stored_by
        sql_stored_by = """
        DELETE FROM stored_by WHERE insertion_id = ?
        """
        self.__execute_sql_qmark(sql_stored_by, (insertion_id, ))

        # backuped_by
        sql_backuped_by = """
        DELETE FROM backuped_by WHERE insertion_id = ?
        """
        self.__execute_sql_qmark(sql_backuped_by, (insertion_id, ))

        # backuped_by
        sql_get_backups = """
        SELECT DISTINCT insertion_id, session_id
        FROM backuped_by
        WHERE insertion_id = ?
        """
        backups = self.__execute_sql_qmark(sql_get_backups, (insertion_id, ))
        for backup in backups:
            # rerank
            self.__rerank_backuped_by(insertion_id, backup[1])
        # remove
        sql_delete_backuped_by = """
        DELETE FROM backuped_by WHERE insertion_id = ?
        """
        self.__execute_sql_qmark(sql_delete_backuped_by, (insertion_id, ))

        # pending_stores
        sql_pending_stores = """
        DELETE FROM pending_stores WHERE insertion_id = ?
        """
        self.__execute_sql_qmark(sql_pending_stores, (insertion_id, ))

        # insertions
        sql_insertions = """
        UPDATE insertions
        SET is_deleted = 1
        WHERE id = ?
        """
        self.__execute_sql_qmark(sql_insertions, (insertion_id, ))

    def store_file(self, insertion_id: str, session_id: str):
        # rerank backuped_by
        self.__rerank_backuped_by(insertion_id, session_id)
        # remove from backuped_by
        sql_delete_backuped_by = """
        DELETE FROM backuped_by WHERE (insertion_id = ?) AND (session_id = ?)
        """
        self.__execute_sql_qmark(sql_delete_backuped_by, (insertion_id, session_id))
        # add to stored_by
        sql_add_to_stored_by = """
        INSERT OR IGNORE INTO stored_by
            (insertion_id, session_id)
        VALUES
            (?, ?)
        """
        self.__execute_sql_qmark(sql_add_to_stored_by, (insertion_id, session_id))

    def set_backups(self, insertion_id: str, backup_list: List[Tuple[str, str]]):
        # remove previous backups
        sql_delete_backuped_by = """
        DELETE FROM backuped_by WHERE (insertion_id = ?)
        """
        self.__execute_sql_qmark(sql_delete_backuped_by, (insertion_id, ))
        # add backups
        length = len(backup_list)
        for rank in range(length):
            backup = backup_list[rank]
            sql_add_backup = """
            INSERT OR IGNORE INTO backuped_by
                (insertion_id, session_id, rank, nonce)
            VALUES
                (?, ?, ?, ?)
            """
            self.__execute_sql_qmark(sql_add_backup, (insertion_id, backup[0], rank, backup[1]))

    def add_backup(self, insertion_id: str, session_id: str, rank: int, nonce: str):
        # delete all backups with larger rank value
        sql_delete_backuped_by = """
        DELETE FROM backuped_by
        WHERE (insertion_id = ?) AND rank >= ?
        """
        self.__execute_sql_qmark(sql_delete_backuped_by, (insertion_id, rank))
        # add this backup
        sql_add_backup = """
        INSERT OR IGNORE INTO backuped_by
            (insertion_id, session_id, rank, nonce)
        VALUES
            (?, ?, ?, ?)
        """
        self.__execute_sql_qmark(sql_add_backup, (insertion_id, session_id, rank, nonce))

    def get_stored_bys(self, insertion_id: str):
        sql = """
        SELECT DISTINCT insertion_id, session_id
        FROM stored_by
        WHERE insertion_id = ?
        ORDER BY session_id ASC
        """
        results = self.__execute_sql_qmark(sql, (insertion_id, ))
        stored_bys = []
        for result in results:
            stored_bys.append(result[1])
        return stored_bys

    def get_backuped_bys(self, insertion_id: str):
        sql = """
        SELECT DISTINCT insertion_id, session_id, rank, nonce
        FROM backuped_by
        WHERE insertion_id = ?
        ORDER BY rank
        """
        results = self.__execute_sql_qmark(sql, (insertion_id, ))
        backuped_bys = []
        for result in results:
            backuped_bys.append({
                'session_id': result[1],
                'rank': result[2],
                'nonce': result[3]
            })
        return backuped_bys

    def get_pending_stores(self, insertion_id: str):
        sql = """
        SELECT DISTINCT insertion_id, session_id
        FROM pending_stores
        WHERE insertion_id = ?
        """
        results = self.__execute_sql_qmark(sql, (insertion_id, ))
        pending_stores = []
        for result in results:
            pending_stores.append(result[1])
        return pending_stores

    def add_pending_store(self, insertion_id: str, session_id: str):
        sql = """
        INSERT OR IGNORE INTO pending_stores
            (insertion_id, session_id)
        VALUES
            (?, ?)
        """
        self.__execute_sql_qmark(sql, (insertion_id, session_id))


# ----------------------------------------------------------
# NDN Hydra Pip Setup
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

#!/usr/bin/env python3
import io
import re
from setuptools import setup, find_packages
from typing import List

with io.open("docs/version.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

with io.open("README.rst", "rt", encoding="utf8") as f:
    long_description = f.read()

def _parse_requirements(filename: str) -> List[str]:
    with open(filename, 'r') as f:
        return [s for s in [ line.split('#', 1)[0].strip(' \t\n') for line in f ] if s != '']

setup(
    name='ndn-hydra',
    version=version,
    description='ndn-hydra: An NDN distributed repository with resiliency coded in python.',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/UCLA-IRL/ndn-hydra',
    author='Justin C Presley',
    author_email='justincpresley@gmail.com',
    maintainer='Justin C Presley',
    maintainer_email='justincpresley@gmail.com',
    download_url='https://pypi.python.org/pypi/ndn-hydra',
    project_urls={
        "Bug Tracker": "https://github.com/UCLA-IRL/ndn-hydra/issues",
        "Source Code": "https://github.com/UCLA-IRL/ndn-hydra",
    },
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Software Development :: Libraries',
        'Topic :: Database',
        'Topic :: Internet',
        'Topic :: System :: Networking',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='NDN HYDRA',
    packages=find_packages(exclude=['tests','examples','docs','notes']),
    install_requires=_parse_requirements('docs/requirements.txt'),
    python_requires=">=3.7",
    entry_points={
        'console_scripts': [
            'ndn-hydra-repo = ndn_hydra.repo.main.main:main',
            'ndn-hydra-client = ndn_hydra.client.main:main'
        ]
    },
    zip_safe=False)


# ----------------------------------------------------------
# NDN Hydra Query Handle
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

import asyncio as aio
import logging
from ndn.app import NDNApp
from ndn.encoding import Name, ContentType, Component
from ndn_python_repo import Storage
from ndn_hydra.repo.global_view.global_view import GlobalView
from ndn_hydra.repo.protocol.repo_commands import File, FileList

class QueryHandle(object):
    """
    QueryHandle processes query interests, and return informational data.
    """
    def __init__(self, app: NDNApp, global_view: GlobalView, config: dict):
        """
        :param app: NDNApp.
        :param global_view: Global View.
        :param config: All config Info.
        """
        self.app = app
        self.global_view = global_view
        self.session_id = config['session_id']
        self.repo_prefix = config['repo_prefix']

        self.logger = logging.getLogger()

        self.command_comp = "/query"
        self.sid_comp = "/sid"

        self.listen(Name.from_str(self.repo_prefix + self.command_comp))
        self.listen(Name.from_str(self.repo_prefix + self.sid_comp  + "/" + self.session_id + self.command_comp))

    def listen(self, prefix):
        """
        This function needs to be called for prefix of all data stored.
        :param prefix: NonStrictName.
        """
        self.app.route(prefix)(self._on_interest)
        self.logger.info(f'Query handle: listening to {Name.to_str(prefix)}')

    def unlisten(self, prefix):
        """
        :param name: NonStrictName.
        """
        aio.ensure_future(self.app.unregister(prefix))
        self.logger.info(f'Query handle: stop listening to {Name.to_str(prefix)}')

    def _on_interest(self, int_name, int_param, _app_param):
        if not int_param.must_be_fresh or not int_param.can_be_prefix:
            return
        query = self._get_query_from_interest(Name.to_str(int_name))
        querytype = Component.to_str(Name.from_str(query)[0])
        if querytype == "sids":
            self.logger.info(f'[cmd][QUERY] query received: sids')
            sessions = self.global_view.get_sessions()
            sidliststr = " ".join([key["id"] for key in sessions])
            self.app.put_data(int_name, content=bytes(sidliststr.encode()), freshness_period=3000, content_type=ContentType.BLOB)
            return
        elif querytype == "files":
            self.logger.info(f'[cmd][QUERY] query received: files')
            insertions = self.global_view.get_insertions()
            filelist = FileList()
            filelist.list = []
            for index in range(len(insertions)):
                file = File()
                file.file_name = insertions[index]["file_name"]
                file.desired_copies = insertions[index]["desired_copies"]
                file.packets = insertions[index]["packets"]
                file.digests = insertions[index]["digests"]
                file.size = insertions[index]["size"]
                filelist.list.append(file)
            self.app.put_data(int_name, content=filelist.encode(), freshness_period=3000, content_type=ContentType.BLOB)
            return
        elif querytype == "file":
            self.logger.info(f'[cmd][QUERY] query received: file')
            insertions = self.global_view.get_insertions()
            filename = query[5:]
            filecontent = None
            for index in range(len(insertions)):
                if Name.to_str(insertions[index]["file_name"]) == filename:
                    file = File()
                    file.file_name = insertions[index]["file_name"]
                    file.desired_copies = insertions[index]["desired_copies"]
                    file.packets = insertions[index]["packets"]
                    file.digests = insertions[index]["digests"]
                    file.size = insertions[index]["size"]
                    filecontent = file.encode()
                    break
            self.app.put_data(int_name, content=filecontent, freshness_period=3000, content_type=ContentType.BLOB)
            return
        elif querytype == "prefix":
            self.logger.info(f'[cmd][QUERY] query received: prefix')
            insertions = self.global_view.get_insertions()
            prefix = query[7:]
            filelist = FileList()
            filelist.list = []
            for index in range(len(insertions)):
                if Name.is_prefix(Name.from_str(prefix), insertions[index]["file_name"]):
                    file = File()
                    file.file_name = insertions[index]["file_name"]
                    file.desired_copies = insertions[index]["desired_copies"]
                    file.packets = insertions[index]["packets"]
                    file.digests = insertions[index]["digests"]
                    file.size = insertions[index]["size"]
                    filelist.list.append(file)
            self.app.put_data(int_name, content=filelist.encode(), freshness_period=3000, content_type=ContentType.BLOB)
            return
        else:
            self.logger.info(f'[cmd][QUERY] unknown query received')
            self.app.put_data(int_name, content=None, freshness_period=3000, content_type=ContentType.NACK)
            return

    def _get_query_from_interest(self, int_name):
        query = int_name[len(self.repo_prefix):]
        if query[0:len(self.sid_comp)] == self.sid_comp:
            return query[(len(self.sid_comp)+len("/" + self.session_id)+len(self.command_comp)):]
        else:
            return query[(len(self.command_comp)):]


# ----------------------------------------------------------
# NDN Hydra Fetch Client
# ----------------------------------------------------------
# @Project: NDN Hydra
# @Date:    2021-01-25
# @Author:  Zixuan Zhong
# @Author:  Justin C Presley
# @Author:  Daniel Achee
# @Source-Code: https://github.com/UCLA-IRL/ndn-hydra
# @Pip-Library: https://pypi.org/project/ndn-hydra/
# ----------------------------------------------------------

from asyncio import Semaphore
import logging
from ndn.app import NDNApp
from ndn.encoding import FormalName, Component, Name, ContentType
import os
from ndn_hydra.repo.utils.concurrent_fetcher import concurrent_fetcher

class HydraFetchClient(object):
    def __init__(self, app: NDNApp, client_prefix: FormalName, repo_prefix: FormalName) -> None:
      """
      This client fetches data packets from the remote repo.
      :param app: NDNApp.
      :param client_prefix: NonStrictName. Routable name to client.
      :param repo_prefix: NonStrictName. Routable name to remote repo.
      """
      self.app = app
      self.client_prefix = client_prefix
      self.repo_prefix = repo_prefix

    async def fetch_file(self, file_name: FormalName, local_filename: str = None, overwrite: bool = False) -> None:
      """
      Fetch a file from remote repo, and write to the current working directory.
      :param name_at_repo: NonStrictName. The name with which this file is stored in the repo.
      :param local_filename: str. The filename of the retrieved file on the local file system.
      :param overwrite: If true, existing files are replaced.
      """
      name_at_repo = self.repo_prefix + [Component.from_str("fetch")] + file_name + [Component.from_segment(0)]

      # If no local filename is provided, store file with last name component
      # of repo filename
      if local_filename is None:
        local_filename = Name.to_str(file_name)
        local_filename = os.path.basename(local_filename)

      # If the file already exists locally and overwrite=False, retrieving the file makes no
      # sense.
      if os.path.isfile(local_filename) and not overwrite:
        raise FileExistsError("{} already exists".format(local_filename))

      # Ping with only one data packet at the start
      b_array = bytearray()
      start_index = 0
      end_index = None
      data_name, meta_info, content, data_bytes = await self.app.express_interest(
        name_at_repo, need_raw_packet=True, can_be_prefix=False, must_be_fresh=False, lifetime=1000)


      # print(content.tobytes())
      if meta_info.content_type == ContentType.NACK:
        print("Distributed Repo does not have that file.")
        return
      elif meta_info.content_type == ContentType.LINK:
        name_at_repo = Name.from_str(bytes(content).decode())
        end_index = Component.to_number(meta_info.final_block_id)
      else:
        # print(type(content))

        name_at_repo = name_at_repo[:-1]
        start_index = start_index + 1
        end_index = Component.to_number(meta_info.final_block_id)

        # print(name_at_repo)
        b_array.extend(content)

      # print(Name.to_str(data_name))

      # Fetch the rest of the file.
      if start_index <= end_index:
          semaphore = Semaphore(10)
          async for (_, _, content, _) in concurrent_fetcher(self.app, Name.from_str("/"), name_at_repo, start_index, end_index, semaphore):
            b_array.extend(content)

      # After b_array is filled, sort out what to do with the data.
      if len(b_array) > 0:
        print(f'Fetching completed, writing to file {local_filename}')

        # Create folder hierarchy
        local_folder = os.path.dirname(local_filename)
        if local_folder:
          os.makedirs(local_folder, exist_ok=True)

        # Write retrieved data to file
        if os.path.isfile(local_filename) and overwrite:
          os.remove(local_filename)
        with open(local_filename, 'wb') as f:
          f.write(b_array)
      else:
        print("Client Fetch Command Failed.")